"""RNAvigate

RNA visualization and graphical analysis toolset

A Jupyter-compatible toolset for visually exploring RNA structure and chemical
probing data.
"""

# from rnavigate.rnavigate import Sample
# from rnavigate.plotting_functions import (
#     plot_options,
#     plot_ss,
# )

# from rnavigate import data
# from rnavigate import plots
# from rnavigate import styles

# __version__ = "1.0.0"
# __author__ = "Patrick S. Irving"
# __email__ = "psirving@email.unc.edu"

# __all__ = [
#     "Sample",
#     # plotting functions
#     "plot_options",
#     "plot_ss",
#     # modules
#     "data",
#     "plots",
#     "styles",
# ]
